import datetime
from pathlib import Path
from re import sub
from string import Template
from typing import Dict


def mk_dates() -> Dict[str, str]:
    today = datetime.datetime.today()
    pretty_date = today.strftime("%b %d, %Y")
    year = today.strftime("%Y")
    return {"DATE_PRETTY": pretty_date, "YEAR": year}


def mk_authors(authors: str) -> Dict[str, str]:
    return {
        "AUTHORS": authors,
        "AUTHORS_SPLIT_NEWLINE": "\n".join(authors.split(", ")),
    }


def to_class_name(s: str) -> str:
    s = sub(r"(_|-)+", " ", s).title().replace(" ", "")
    return "".join([s[0].upper(), s[1:]])


def write_templated_file(
    template_path: Path,
    target_fname: str,
    substitutions: Dict[str, str],
    path_to_target: Path,
) -> None:
    with open(template_path, "r") as f:
        template = Template(f.read())

    with open(path_to_target / Path(target_fname), "w") as f:
        f.write(template.substitute(**substitutions))



import sys
from pathlib import Path

from qtpy import QtGui, QtWidgets

from ScopeFoundry.tools.page import Page
from ScopeFoundry.tools.pages.new_app import NewApp
from ScopeFoundry.tools.pages.new_hardware import NewHardware
from ScopeFoundry.tools.pages.new_measurement import NewMeasurement
from ScopeFoundry.tools.pages.welcome import WelcomePage

ROOT = Path(__file__).parent.parent
ICONS_PATH = ROOT / "base_app/icons"


class ToolsApp(QtWidgets.QMainWindow):

    def __init__(self) -> None:
        super().__init__()

        self.tab_widget = QtWidgets.QTabWidget()
        self.setCentralWidget(self.tab_widget)
        self.pages = []

    def add_page(self, page: Page) -> None:
        page.setup()
        page.setup_figure()
        self.tab_widget.addTab(page.ui, page.name)
        self.pages.append(page.name)

    def show_page(self, name: str) -> None:
        self.tab_widget.setCurrentIndex(self.pages.index(name))


def main() -> None:
    if len(sys.argv) > 1:
        start_app(" ".join(sys.argv[1:]))
    else:
        start_app()


def start_app(page="Welcome") -> None:
    app = QtWidgets.QApplication([])
    window = ToolsApp()
    window.setWindowTitle("ScopeFoundry tools")
    logo_icon = QtGui.QIcon(str(ICONS_PATH / "scopefoundry_logo2C_1024.png"))
    window.setWindowIcon(logo_icon)
    window.add_page(WelcomePage())
    window.add_page(NewApp())
    window.add_page(NewHardware())
    window.add_page(NewMeasurement())
    window.show()
    window.show_page(page)
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



from ScopeFoundry import HardwareComponent
from ScopeFoundry.examples.ScopeFoundryHW.simulon_xyzstage.simulon_xyz_stage_dev import (
    SimulonXYZStageDev,
)


class SimulonXYZStageHW(HardwareComponent):

    name = "simulon_xyz_stage"

    def setup(self):
        position_params = {
            "dtype": float,
            "ro": False,
            "initial": -1,
            "vmin": -1,
            "vmax": 100,
            "si": False,
            "unit": "um",
            "reread_from_hardware_after_write": True,
            "spinbox_decimals": 3,
        }

        self.x_position = self.settings.New("x_position", **position_params)
        self.y_position = self.settings.New("y_position", **position_params)
        self.z_position = self.settings.New("z_position", **position_params)

    def connect(self):

        dev = self.stage_device = SimulonXYZStageDev(debug=self.debug_mode.val)

        self.x_position.connect_to_hardware(dev.read_x, dev.write_x)
        self.y_position.connect_to_hardware(dev.read_y, dev.write_y)
        self.z_position.connect_to_hardware(dev.read_z, dev.write_z)

    def disconnect(self):

        self.settings.disconnect_all_from_hardware()

        if hasattr(self, "stage_device"):
            self.stage_device.close()
            del self.stage_device



import unittest
from ScopeFoundry.logged_quantity import LoggedQuantity


INITIAL_CHOICES = [("choice 1", 1), ("choice 2", 2)]


class ChoiceListTest(unittest.TestCase):

    def setUp(self):
        self.lq = LoggedQuantity("choices", int, choices=INITIAL_CHOICES)

    def test_initial_value_set(self): 
        self.assertEqual(self.lq.val, 1)

    def test_change_choice_list_all_new_choices(self): 
        self.lq.change_choice_list([("choice 3", 3)])
        self.assertEqual(self.lq.val, 3)

    def test_change_choice_list_containing_oldval(self): 
        self.lq.change_choice_list([("choice 1 renamed", 1), ("choice 3", 3)])
        self.assertEqual(self.lq.val, 1)

    def test_change_choice_list_all_new(self): 
        self.lq.change_choice_list([("choice 3", 3), ("choice 4", 4)])
        self.assertEqual(self.lq.val, 3)

    def test_change_choice_list_empty(self): 
        self.lq.change_choice_list([])
        self.assertEqual(self.lq.val, 1)  # for now expected behavior

    def test_add_choices(self): 
        self.lq.add_choices([("choice 4", 4)])
        self.assertEqual(self.lq.val, 1)
        self.assertIn(("choice 4", 4), self.lq.choices)

    def test_add_choices_overlapping(self): 
        self.lq.add_choices([("choice 2", 2), ("choice 4", 4)])
        self.assertEqual(self.lq.val, 1)
        self.assertIn(("choice 1", 1), self.lq.choices)
        self.assertIn(("choice 2", 2), self.lq.choices)
        self.assertIn(("choice 4", 4), self.lq.choices)

    def test_add_choices_and_set(self): 
        self.lq.add_choices([("choice 4", 4)], new_val=2)
        self.assertEqual(self.lq.val, 2)

    def test_add_empty(self): 
        self.lq.add_choices([])
        self.assertEqual(self.lq.val, 1)  # for now expected behavior

    def test_remove_a_set_choice(self): 
        self.lq.remove_choices([("choice 1", 1)])
        self.assertEqual(self.lq.val, 2)

    def test_remove_unset_choice(self): 
        self.lq.remove_choices([("choice 2", 2)])
        self.assertEqual(self.lq.val, 1)

    def test_remove_all(self): 
        self.lq.remove_choices(INITIAL_CHOICES)
        self.assertEqual(self.lq.val, 1)  # for now expected behavior


if __name__ == '__main__':
    unittest.main()



import sys


from ScopeFoundry import BaseMicroscopeApp
from ScopeFoundry.examples.example_2d_slowscan_measure import Example2DSlowScanMeasure
from ScopeFoundry.examples.example_3d_slowscan_measure import Example3DSlowScanMeasure
from ScopeFoundry.examples.ScopeFoundryHW.bsinc_noiser200.bsinc_noiser200_hw import (
    Noiser200HW,
)
from ScopeFoundry.examples.ScopeFoundryHW.simulon_xyzstage.simulon_xyz_stage_hw import (
    SimulonXYZStageHW,
)


class Example2DSlowScanApp(BaseMicroscopeApp):

    name = "Example 2D Slow Scan"

    def setup(self):

        from ScopeFoundry.sequencer import Sequencer

        self.add_measurement(Sequencer)

        self.add_hardware(SimulonXYZStageHW(self))
        self.add_hardware(Noiser200HW(self))

        self.add_measurement(Example2DSlowScanMeasure(self))
        self.add_measurement(Example3DSlowScanMeasure(self))


if __name__ == "__main__":
    app = Example2DSlowScanApp(sys.argv)
    app.settings_load_ini("default_settings.ini")
    sys.exit(app.exec_())



from qtpy import QtCore


class LQCircularNetwork(QtCore.QObject):
    """
    LQCircularNetwork is collection of logged quantities
    Helper class if a network of lqs with circular structures are present
    Use update_values_synchronously method if the specified lqs should
    be updated at once. The internal lock-flag prevents infinite loops.
    """

    updated_values = QtCore.Signal(
        (),
    )

    def __init__(self, lq_dict=None, lq_list=None):
        if lq_dict is None:
            lq_dict = dict()
        if lq_list is not None:
            for lq in lq_list:
                lq_dict[lq.name] = lq
        self.lq_dict = lq_dict  # {lq_key:lq}
        self.locked = False  # some lock that does NOT allow blocked routines to be executed after release()
        # a flag (as it is now) works well.

    def update_values_synchronously(self, **kwargs):
        """
        kwargs is dict containing lq_key and new_vals
        Note: lq_key is not necessarily the name of the lq but key of lq_dict
              as specified at initialization
        """
        if self.locked == False:
            self.locked = True
            for kev, val in kwargs.items():
                self.lq_dict[kev].update_value(val)
                self.updated_values.emit()
                self.locked = False

    def add_lq(self, lq, name=None):
        if name is None:
            name = lq.name
        self.lq_dict[name] = lq



"""
Created on Tue Apr  1 09:25:48 2014
@author: esbarnard
"""

import sys
import threading
import time
import traceback
from typing import Callable
from warnings import warn

from qtpy import QtCore, QtGui, QtWidgets

from .operations import Operations
from .base_app import BaseMicroscopeApp
from .dynamical_widgets.generic_widget import add_to_layout, new_widget
from .dynamical_widgets.tree_widget import SubtreeManager
from .helper_funcs import get_logger_from_class, load_qt_ui_file
from .logged_quantity import LQCollection


class MeasurementQThread(QtCore.QThread):
    def __init__(self, measurement, parent=None):
        super(MeasurementQThread, self).__init__(parent)
        self.measurement = measurement

    def run(self):
        self.measurement._thread_run()


class Measurement:
    """
    Base class for ScopeFoundry Measurement objects

    to subclass, implement :meth:`setup`, :meth:`run`

    for measurements with graphical interfaces,
    subclass and additionally implement :meth:`setup_figure`, :meth:`update_display`


    Run States:

    stop_first -> run_starting -> run_pre_run --> run_thread_starting --> run_thread_run -->

    run_thread_end --> run_post_run --> stop_success | stop_interrupted | stop_failure

    """

    def __init__(self, app: BaseMicroscopeApp, name: str = None):

        self.q_object = MeasurementQObject(self)
        self.measurement_sucessfully_completed = (
            self.q_object.measurement_sucessfully_completed
        )
        self.measurement_interrupted = self.q_object.measurement_interrupted

        self.log = get_logger_from_class(self)

        if not hasattr(self, "name"):
            self.name = self.__class__.__name__

        if name is not None:
            self.name = name

        self.app = app

        self.display_update_period = 0.1  # seconds

        self.acq_thread = None

        self.interrupt_measurement_called = False

        self.settings = LQCollection(path=f"mm/{self.name}")
        self.operations = Operations()
        self._subtree_managers_ = []
        self._widgets_managers_ = []

        self.activation = self.settings.New(
            name="activation",
            dtype=bool,
            ro=False,
            description=f"<i>{self.name}</i>",
            protected=True,
        )  # does the user want to the thread to be running

        self.run_state = self.settings.New(
            "run_state", dtype=str, initial="stop_first", ro=True, protected=True
        )
        self.progress = self.settings.New(
            "progress", dtype=float, unit="%", si=False, ro=True, protected=True
        )
        self.settings.New(
            "profile",
            dtype=bool,
            initial=False,
            description="Run a profile on the run to find performance problems",
        )

        self.activation.updated_value[bool].connect(self.start_stop)

        self.add_operation("start", self.start)
        self.add_operation("interrupt", self.interrupt)
        # self.add_operation('terminate', self.terminate)
        # self.add_operation("setup", self.setup)
        # self.add_operation("setup_figure", self.setup_figure)
        self.add_operation("update_display", self.update_display)
        self.add_operation(
            name="show_ui",
            op_func=self.show_ui,
            icon_path=self.app.icons_path / "show_ui_logo.png",
        )
        self.add_operation("Reload_Code", self.reload_code)

        if hasattr(self, "ui_filename"):
            self.load_ui()

        self.setup()

    def setup(self):
        """Override this to set up logged quantities and gui connections
        Runs during __init__, before the hardware connection is established
        Should generate desired LoggedQuantities"""
        pass
        # raise NotImplementedError()

    def setup_figure(self):
        """
        Override setup_figure to build graphical interfaces.
        This function is run on ScopeFoundry startup.
        """
        self.log.info("Empty setup_figure called")
        pass

    def start(self):
        """
        Starts the measurement

        calls *pre_run*
        creates acquisition thread
        runs thread
        starts display timer which calls update_display periodically
        calls post run when thread is finished
        """
        # self.start_stop(True)
        self.activation.update_value(True)

    def _start(self):
        """
        INTERNAL DO NOT CALL DIRECTLY

        Starts the measurement

        calls *pre_run*
        creates acquisition thread
        runs thread
        starts display timer which calls update_display periodically
        connects a signal/slot that calls post run when thread is finished
        """
        self.interrupt_measurement_called = False
        self.run_state.update_value("run_starting")

        msg = f"measurement {self.name} start called from thread: {repr(threading.get_ident())}"
        self.log.info(msg)

        if self.is_thread_alive():
            msg = f"Cannot start a new measurement while still measuring {self.acq_thread} {self.is_measuring()}"
            raise RuntimeError(msg)

        # remove previous qthread with delete later
        # if self.acq_thread is not None:
        #    self.acq_thread.deleteLater()
        self.acq_thread = MeasurementQThread(self)
        self.acq_thread.finished.connect(self._call_post_run)
        # self.measurement_state_changed.emit(True)
        # self.running.update_value(True)
        self.run_state.update_value("run_prerun")
        try:
            self.pre_run()
        except Exception as err:
            # print("err", err)
            self.run_state.update_value("stop_failure")
            self.activation.update_value(False)
            raise

        self.run_state.update_value("run_thread_starting")
        self.acq_thread.start()
        self.run_state.update_value("run_thread_run")
        self.t_start = time.time()
        self.q_object.display_update_timer.start(int(self.display_update_period * 1000))

    def pre_run(self):
        """Override this method to enable main-thread initialization prior to measurement thread start"""
        pass

    def run(self):
        """
        *run* method runs in an separate thread and is used for data acquisition

        No GUI updates should occur within the *run* function, any Qt related GUI work
        should occur in :meth:`update_display`

        Don't call this directly!
        """

        if hasattr(self, "_run"):
            self.log.warning("warning _run is deprecated, use run")
            self._run()
        else:
            raise NotImplementedError(f"Measurement {self.name}.run() not defined")

    def _call_post_run(self):
        """
        Don't call this directly!
        """
        self.run_state.update_value("run_post_run")
        try:
            self.post_run()
        except Exception as err:
            raise
        finally:
            self.activation.update_value(False)
            self.run_state.update_value(self.end_state)

    def post_run(self):
        """Override this method to enable main-thread finalization after to measurement thread completes"""
        pass

    def _thread_run(self):
        """
        This function governs the behavior of the measurement thread.
        """
        print(self.name, "_thread_run thread_id:", threading.get_ident())
        self.progress.update_value(50.0)  # default 50% w/o time remaining estimation
        self._t0 = time.time()
        try:
            if self.settings["profile"]:
                import cProfile

                profile = cProfile.Profile()
                profile.enable()
            self.run()
            success = True
        except Exception as err:
            success = False
            raise
        finally:
            self.run_state.update_value("run_thread_end")

            # self.running.update_value(False)
            self.set_progress(0.0)  # set progress bars back to zero
            # self.measurement_state_changed.emit(False)
            if self.interrupt_measurement_called:
                self.measurement_interrupted.emit()
                self.interrupt_measurement_called = False
                end_state = "stop_interrupted"
            elif not success:
                end_state = "stop_failure"
            else:
                self.measurement_sucessfully_completed.emit()
                end_state = "stop_success"
            if self.settings["profile"]:
                profile.disable()
                profile.print_stats(sort="time")

            self.end_state = end_state

    @property
    def gui(self):
        self.log.warning(
            "Measurement.gui is deprecated, use Measurement.app "
            + repr(DeprecationWarning)
        )
        return self.app

    def set_progress(self, pct):
        """
        This function updates the logged quantity progress which is used for the display of progress bars in the UI.

        ==============  ==============================================================================================
        **Arguments:**
        pct             The percentage of progress given by a measurement module
        ==============  ==============================================================================================
        """
        self.progress.update_value(pct)

        if pct:
            text = f"{self.name} (in {to_etr_str((100 - pct)/pct * (time.time() - self._t0))})"
        else:
            text = self.name

        if hasattr(self, "subwin"):
            self.subwin.setWindowTitle(text)

        for manager in self._subtree_managers_:
            manager.set_header_text(0, text, None)

    def _interrupt(self):
        """
        Kindly ask the measurement to stop.

        This raises the :attr:`interrupt_measurement_called` flag
        To actually stop, the threaded :meth:`run` method must check
        for this flag and exit
        """
        self.log.info(f"measurement {self.name} stopping {self.settings['run_state']}")
        # print("{self.name} interrupt(): run_state={ self.settings['run_state']}")
        if self.settings["run_state"].startswith("run"):
            self.log.info(f"measurement {self.name} interrupt called")
            self.interrupt_measurement_called = True
        # self.activation.update_value(False)
        # Make sure display is up to date
        # self.q_object._on_display_update_timer()

    def interrupt(self):
        self.activation.update_value(False)

    def terminate(self):
        """
        Terminate MeasurementQThread. Usually a bad idea:
        This will not clean up the thread correctly and usually
        requires a reboot of the App
        """
        self.acq_thread.terminate()

    def start_stop(self, start):
        """
        Use boolean *start* to either start (True) or
        interrupt (False) measurement. Test.
        """
        self.log.info(f"{self.name} start_stop {start}")
        if start:
            self._start()
        else:
            self._interrupt()

    def is_measuring(self):
        """
        Returns whether the acquisition thread is running
        """
        # print(self.name, "is_measuring run_state", self.settings['run_state'])
        return self.settings["run_state"].startswith("run")

        """
        if self.acq_thread is None:
            self.running.update_value(False)
            self.activation.update_value(False)
            self.settings['progress'] = 0.0
            return False
        else:
            #resp =  self.acq_thread.is_alive()
            resp = self.acq_thread.isRunning()
            self.running.update_value(resp)            
            return resp
        """

    def is_thread_alive(self):
        if self.acq_thread is None:
            return False
        else:
            # resp =  self.acq_thread.is_alive()
            resp = self.acq_thread.isRunning()
            return resp

    def update_display(self):
        "Override this function to provide figure updates when the display timer runs"
        pass

    def add_logged_quantity(self, name, **kwargs):
        """
        Create a new :class:`LoggedQuantity` and adds it to the measurement's
        :attr:`settings` (:class:`LQCollection`)
        """
        lq = self.settings.New(name=name, **kwargs)
        return lq

    def start_nested_measure_and_wait(
        self,
        measure,
        nested_interrupt: bool = True,
        polling_func: Callable = None,
        polling_time: float = 0.1,
    ):
        """
        Start another nested measurement *measure* and wait until completion.
        Should be called with run function.


        Optionally it can call a polling function *polling_func* with no arguments
        at an interval *polling_time* in seconds.

        if *nested_interrupt* is True then interrupting the nested *measure* will
        also interrupt the outer measurement. *nested_interrupt* defaults to True

        returns True if successful run, otherwise returns false for a run failure or interrupted measurement
        """

        self.log.info(
            f"Starting nested measurement {measure.name} from {self.name} on thread id {threading.get_ident()}"
        )

        measure.interrupt_measurement_called = False
        measure.start()

        # Wait until measurement has started, timeout of 1 second
        t0 = time.time()
        while not measure.is_measuring():
            time.sleep(0.010)
            if time.time() - t0 > 1.0:
                print(
                    f"{self.name}: nested measurement {measure.name} has not started before timeout"
                )
                return measure.settings["run_state"] == "stop_success"

        last_polling = time.time()

        # Now that it is running, wait until done
        while measure.is_measuring():
            if self.interrupt_measurement_called:
                # print('nest outer interrupted', self.interrupt_measurement_called)
                measure.interrupt()

            if measure.interrupt_measurement_called and nested_interrupt:
                # THIS IS MAYBE UNSAFE???: measure.interrupt_measurement_called might be also TRUE if measure finished successfully?
                # IDEA to TEST: also check the measure.settings['run_state'].startswidth('stop')
                print(
                    "nested interrupt bubbling up",
                    measure.interrupt_measurement_called,
                    self.interrupt_measurement_called,
                )
                self.interrupt()

            time.sleep(0.010)

            # polling
            t = time.time()
            if (
                measure.settings["run_state"] == "run_thread_run"
                and polling_func
                and t - last_polling > polling_time
            ):
                try:
                    polling_func()
                except Exception as err:
                    self.log.error(
                        f"start_nested_measure_and_wait polling failed {err}"
                    )
                last_polling = t

        # returns True if successful run, otherwise,
        # returns false for a run failure or interrupted measurement
        return measure.settings["run_state"] == "stop_success"

    def load_ui(self, ui_fname=None):
        """
        Loads and shows user interface.

        ==============  ===============================================================
        **Arguments:**
        ui_fname        filename of user interface file (usually made with Qt Designer)
        ==============  ===============================================================
        """

        # TODO destroy and rebuild UI if it already exists
        if ui_fname is not None:
            self.ui_filename = ui_fname
        # Load Qt UI from .ui file
        self.ui = load_qt_ui_file(self.ui_filename)
        # self.show_ui()

    def show_ui(self):
        """
        Shows the graphical user interface of this measurement. :attr:`ui`
        """
        self.app.bring_measure_ui_to_front(self)

    def new_control_widgets(self, title: str = None, include=None, exclude=None):
        """creates scroll area group box that updates on dynamical add/remove of settings/operations"""
        if title is None:
            title = self.name
        return new_widget(self, title, include, exclude)

    def add_to_layout(self, layout, include=None, exclude=None):
        add_to_layout(self, layout, include, exclude)

    def new_start_stop_button(self, texts=("▶  START", "🛑  STOP")):
        btn = QtWidgets.QPushButton()
        btn.setObjectName("start_stop_button")
        self.activation.connect_to_pushButton(btn, (), texts, "")
        if len(texts[0]) <= 1 and len(texts[1]) <= 1:
            btn.setFixedWidth(22)
        return btn

    def add_operation(
        self, name: str, op_func: Callable[[], None], description="", icon_path=""
    ):
        """
        Create an operation for the Measurement.

        *op_func* is a function that will be called upon operation activation

        operations are typically exposed in the default ScopeFoundry gui via a pushButton

        :type name: str
        :type op_func: QtCore.Slot or Callable without Argument
        :type description: str
        """
        self.operations.new(name, op_func, description, icon_path)

    def remove_operation(self, name: str):
        self.operations.remove(name)

    def web_ui(self):
        return f"Hardware {self.name}"

    def reload_code(self):
        import inspect

        import xreload

        mod = inspect.getmodule(self)
        x = xreload.xreload(mod)
        print("Reloading from code", mod, x)

    def on_new_subtree(self, subtree: SubtreeManager):
        progress_bar = QtWidgets.QProgressBar()
        progress_bar.setObjectName("measurement_progress")
        self.progress.connect_to_widget(progress_bar)

        btn = self.new_start_stop_button(texts=("▶", "🛑"))
        btn.clicked.connect(self.show_ui)

        widget = QtWidgets.QWidget()
        layout = QtWidgets.QHBoxLayout(widget)
        layout.addWidget(progress_bar)
        layout.addWidget(btn)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        subtree.tree_widget.setItemWidget(subtree.header_item, 1, widget)

        subtree.progress_bar = progress_bar

    @property
    def tree_progressBar(self):
        # for backward compatibility
        warn(
            "Measurement.tree_progressBar deprecated, loop through Measurement._subtree_managers_[i].progress_bar instead.",
            DeprecationWarning,
        )
        return self._subtree_managers_[0].progress_bar

    def on_right_click(self):
        cmenu = QtWidgets.QMenu()
        cmenu.addAction(self.name).setEnabled(False)
        cmenu.addSeparator()
        cmenu.addAction("Start", self.start)
        cmenu.addAction("Interrupt", self.interrupt)
        cmenu.addSeparator()
        ui = self.app.load_measure_ui(self)
        if ui is not None:
            cmenu.addAction("Show", self.show_ui)
        cmenu.exec_(QtGui.QCursor.pos())


class MeasurementQObject(QtCore.QObject):

    measurement_sucessfully_completed = QtCore.Signal(())
    """signal sent when full measurement is complete"""
    measurement_interrupted = QtCore.Signal(())
    """signal sent when  measurement is complete due to an interruption"""

    def __init__(self, measurement: Measurement, parent: QtCore.QObject = None) -> None:
        super().__init__(parent)
        self.m = measurement
        self.display_update_timer = QtCore.QTimer()
        self.display_update_timer.timeout.connect(self._on_display_update_timer)

    @QtCore.Slot()
    def _on_display_update_timer(self):
        try:
            self.m.update_display()
        except Exception as err:
            tb = "\n".join(traceback.format_exception(*sys.exc_info()))
            self.m.log.error(f"{self.m.name} failed to update display: {err}. {tb}")
        finally:
            if not self.m.is_measuring():
                self.display_update_timer.stop()


def to_etr_str(secs):
    mins, mins_rem = divmod(secs, 60)
    if not mins:
        return f"{int(secs)}s"
    hours, hours_rem = divmod(secs, 3600)
    if not hours:
        return f"{mins + mins_rem/60:0.1f}min"
    days, days_rem = divmod(secs, 86400)
    if not days:
        return f"{hours + hours_rem/3600:0.1f}h"
    return f"{secs/86400:0.1f}d"



from qtpy.QtCore import QObject
from qtpy.QtWidgets import QGroupBox, QHBoxLayout, QPushButton, QSizePolicy, QWidget
from qtpy.QtCore import QEvent


class MyGroupBox(QGroupBox):
    def __init__(self, title, parent=None):
        super().__init__(title, parent)

        self.setObjectName("sequencer_editor")
        self.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Maximum)

    def hook_children_widget_focus(self):
        for widget in self.children():
            widget.installEventFilter(self)

    def eventFilter(self, watched: QObject, event: QEvent) -> bool:
        if event.type() == QEvent.FocusIn:
            self.setStyleSheet("QGroupBox {border: 1px solid rgba(105, 240, 104, 220)}")
        if event.type() == QEvent.FocusOut:
            self.setStyleSheet("QGroupBox {border: none}")
        return watched.eventFilter(watched, event)


class EditorBaseUI:
    item_type = ""
    description = ""

    def __init__(self, measure) -> None:
        self.measure = measure

        self.group_box = gb = MyGroupBox(
            self.item_type.replace("_", " ").replace("-", " ").title()
        )
        gb.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Maximum)
        gb.setToolTip(self.description)
        self.layout = layout = QHBoxLayout(gb)

        self.add_btn = add_btn = QPushButton("New")
        add_btn.setToolTip("CTRL+N")
        add_btn.setFixedWidth(33)
        layout.addWidget(add_btn)

        self.replace_btn = replace_btn = QPushButton("Replace")
        replace_btn.setFixedWidth(50)
        replace_btn.setToolTip("CTRL+R")
        layout.addWidget(replace_btn)

        self.setup_ui()
        gb.hook_children_widget_focus()

    def set_on_new_func(self, fn):
        self.add_btn.clicked.connect(fn)

    def set_on_replace_func(self, fn):
        self.replace_btn.clicked.connect(fn)

    def setup_ui(self): ...

    def get_kwargs(self): ...

    def set_kwargs(self, **kwargs): ...



import sys

from ScopeFoundry.data_browser import DataBrowser
from ScopeFoundry.data_browser.plug_ins.h5_search import H5SearchPlugIn
from ScopeFoundry.data_browser.plug_ins.time_note import TimeNote
from ScopeFoundry.data_browser.viewers import H5TreeView, RangedOptimizationH5View
from ScopeFoundry.helper_funcs import sibling_path

LOGO_PATH = sibling_path(__file__, "logo.png")


class MyDataBrowser(DataBrowser):

    def setup(self):

        self.add_plugin(H5SearchPlugIn(self))
        self.add_plugin(TimeNote(self))

        self.add_view(H5TreeView(self))
        self.add_view(RangedOptimizationH5View(self))


if __name__ == "__main__":
    app = MyDataBrowser(sys.argv)
    app.set_logo(LOGO_PATH)
    app.settings_load_ini("defaults.ini")

    sys.exit(app.exec_())



import unittest

from ScopeFoundry.logged_quantity.collection import LQCollection


class LQRangeTest(unittest.TestCase):

    def setUp(self):
        self.settings = LQCollection("")
        self.settings.New_Range("x", True, False, (1000.0, 2000.0, 51))

    def test_set_initials(self):

        s = LQCollection("")
        s.New_Range("x", True, False, (1000.0, 2000.0, 100))

        self.assertEqual(s["x_min"], 1000.0)
        self.assertEqual(s["x_max"], 2000.0)
        self.assertEqual(s["x_center"], 1500.0)
        self.assertEqual(s["x_num"], 11)
        self.assertEqual(s["x_step"], 100.0)
        self.assertEqual(s["x_span"], 1000.0)

    def test_set_bounds(self):
        s = self.settings

        s["x_num"] = 11
        s["x_min"] = 0
        s["x_max"] = 1

        self.assertEqual(s["x_center"], 0.5)

    def test_update_center(self):
        s = self.settings

        s["x_num"] = 11
        s["x_min"] = 0
        s["x_max"] = 1
        s["x_center"] = 0.7

        self.assertAlmostEqual(s["x_min"], 0.2)
        self.assertAlmostEqual(s["x_max"], 1.2)

    def test_update_center_span(self):
        s = self.settings

        s["x_num"] = 11
        s["x_min"] = 0
        s["x_max"] = 1

        s["x_center"] = 0
        s["x_span"] = 100

        self.assertEqual(s["x_min"], -50)
        self.assertEqual(s["x_max"], 50)

    def test_span_num(self):
        s = self.settings

        s["x_span"] = 200
        s["x_num"] = 101

        self.assertEqual(s["x_step"], 2)


if __name__ == "__main__":
    unittest.main()



import unittest

from ScopeFoundry.tests.unittests.get_settings import GetSettingsTest
from ScopeFoundry.tests.unittests.lq_choice_list_test import ChoiceListTest
from ScopeFoundry.tests.unittests.settings_io_test import SettingsIOTest
from ScopeFoundry.tests.unittests.lq_connection_test import LQConnectionTest
from ScopeFoundry.tests.unittests.lq_range_test import LQRangeTest
from ScopeFoundry.tests.unittests.analyze_nb_test import AnalyzeNBTest
from ScopeFoundry.tests.unittests.test_operations import TestOperations


# following also require visual inspection - run individual files
from ScopeFoundry.tests.hw_connect_failure_test import TestFailHW
from ScopeFoundry.tests.nested_measurement_test import NestMeasureTestAppTest
from ScopeFoundry.tests.quickbar_test import QuickbarTest

if __name__ == "__main__":
    unittest.main()



from collections import OrderedDict
from typing import Dict, List
from warnings import warn

from qtpy import QtCore, QtWidgets

from ScopeFoundry.dynamical_widgets.tools import Tools
from ScopeFoundry.helper_funcs import get_logger_from_class, filter_with_patterns
from ScopeFoundry.logged_quantity import (
    ArrayLQ,
    FileLQ,
    LoggedQuantity,
    LQ3Vector,
    LQRange,
)


class LQCollectionQObject(QtCore.QObject):

    lq_added = QtCore.Signal((LoggedQuantity,))
    lq_removed = QtCore.Signal((LoggedQuantity,))

    def __init__(self, settings, parent: QtCore.QObject = None) -> None:
        super().__init__(parent)
        self.settings = settings


class LQCollection:
    """
    LQCollection is a smart dictionary of LoggedQuantity objects.

    attribute access such as lqcoll.x1 will return full LoggedQuantity object

    dictionary-style access lqcoll['x1'] allows direct reading and writing of
    the LQ's value, handling the signals automatically

    New LQ's can be created with :meth:`New`

    LQRange objects can be created with :meth:`New_Range` and will be stored
    in :attr:ranges

    """

    def __init__(self, path="") -> None:
        self._logged_quantities = OrderedDict()
        self.ranges = OrderedDict()
        self.vectors = OrderedDict()
        self.log = get_logger_from_class(self)
        self.q_object = LQCollectionQObject(self)
        self.path = path
        self._widgets_managers_ = []

    def new_file(
        self,
        name: str,
        initial: float = 0.0,
        is_dir: bool = False,
        default_dir: str = None,
        file_filters=(),
        description: str = None,
        ro: bool = False,  # read only flag
        choices=None,
        colors=None,
        protected: bool = False,  # a guard that prevents from being updated, i.e. file loading
        is_clipboardable=False,
        **kwargs,
    ) -> FileLQ:
        return self.New(
            name=name,
            dtype="file",
            initial=initial,
            is_dir=is_dir,
            default_dir=default_dir,
            file_filters=file_filters,
            description=description,
            ro=ro,
            choices=choices,
            colors=colors,
            protected=protected,
            is_clipboardable=is_clipboardable,
            **kwargs,
        )

    def New(
        self,
        name: str,
        dtype: type = float,
        initial: float = 0.0,
        unit: str = None,
        si: bool = False,
        vmin: float = -1_000_000_000_000,
        vmax: float = 1_000_000_000_000,
        description: str = None,
        ro: bool = False,  # read only flag
        choices=None,
        spinbox_decimals: int = 2,
        spinbox_step: float = 0.1,
        colors=None,
        protected: bool = False,  # a guard that prevents from being updated, i.e. file loading
        is_dir: bool = False,
        default_dir: str = None,
        file_filters=(),
        reread_from_hardware_after_write: bool = False,
        is_array: bool = False,
        fmt="%g",
        is_cmd=False,
        is_clipboardable=False,
        default_widget_factory=None,
        **kwargs,
    ) -> LoggedQuantity:
        """
        Create a new LoggedQuantity with name and dtype
        """
        if "array" in kwargs:
            warn(
                f"New(array=...) is deprecated, use New(is_array=...) instead. Seen in creation of setting {name}",
                DeprecationWarning,
            )
            is_array = is_array or kwargs.pop("array", False)

        kwargs.update(
            {
                "initial": initial,
                "unit": unit,
                "si": si,
                "ro": ro,
                "choices": choices,
                "vmin": vmin,
                "vmax": vmax,
                "fmt": fmt,
                "description": description,
                "protected": protected,
            }
        )

        if is_array:
            lq = ArrayLQ(name=name, dtype=dtype, **kwargs)
        elif dtype == "file":
            lq = FileLQ(
                name=name,
                default_dir=default_dir,
                is_dir=is_dir,
                file_filters=file_filters,
                is_clipboardable=is_clipboardable,
                **kwargs,
            )
        else:
            lq = LoggedQuantity(
                name=name,
                dtype=dtype,
                spinbox_decimals=spinbox_decimals,
                spinbox_step=spinbox_step,
                reread_from_hardware_after_write=reread_from_hardware_after_write,
                colors=colors,
                is_cmd=is_cmd,
                is_clipboardable=is_clipboardable,
                default_widget_factory=default_widget_factory,
                **kwargs,
            )

        return self.Add(lq)

    def Add(self, lq: LoggedQuantity) -> LoggedQuantity:
        """Add an existing LoggedQuantity to the Collection
        Examples of usefulness: add hardware lq to measurement settings
        """
        name = lq.name
        assert not (name in self._logged_quantities)
        assert not (name in self.__dict__)
        self._logged_quantities[name] = lq
        self.__dict__[name] = lq  # allow attribute access

        lq.set_path(f"{self.path}/{name}")
        self.q_object.lq_added[LoggedQuantity].emit(lq)
        return lq

    def get_lq(self, key) -> LoggedQuantity:
        return self._logged_quantities[key]

    def get_val(self, key):
        return self._logged_quantities[key].val

    def as_list(self) -> List[LoggedQuantity]:
        return self._logged_quantities.values()

    def as_dict(self) -> Dict[str, LoggedQuantity]:
        return self._logged_quantities

    def as_value_dict(self):
        val_dict = dict()
        for k, lq in self.as_dict().items():
            val_dict[k] = lq.value
        return val_dict

    #    def items(self):
    #        return self._logged_quantities.items()

    def keys(self):
        return self._logged_quantities.keys()

    def remove(self, name):
        lq = self._logged_quantities.pop(name)
        self.q_object.lq_removed.emit(lq)
        del self.__dict__[name]
        del lq

    def __delitem__(self, key):
        self.remove(key)

    def __getitem__(self, key):
        "Dictionary-like access reads and sets value of LQ's"
        return self._logged_quantities[key].val

    def __setitem__(self, key, item):
        "Dictionary-like access reads and sets value of LQ's"
        self._logged_quantities[key].update_value(item)

    def __contains__(self, key):
        return self._logged_quantities.__contains__(key)

    def __str__(self) -> str:
        lines = [f"Settings path: {self.path}"] + [
            f"{k}: {v}" for k, v in self.as_value_dict().items()
        ]
        return "\n".join(lines)

    """
    def __getattribute__(self,name):
        if name in self.logged_quantities.keys():
            return self.logged_quantities[name]
        else:
            return object.__getattribute__(self, name)
    """

    def New_Range(
        self,
        name: str,
        include_center_span: bool = False,
        include_sweep_type: bool = False,
        initials=[0, 1.0, 0.1],
        unit: str = "",
        si: bool = False,
        ro: bool = False,
        vmin: float = -1_000_000_000_000,
        vmax: float = 1_000_000_000_000,
        spinbox_decimals: int = 2,
        description: str = "",
        **kwargs,
    ):
        kwargs["ro"] = ro
        kwargs["description"] = description

        kwargs.pop("dtype", None)

        mn, mx, d = initials
        min_lq = self.New(
            f"{name}_min",
            initial=mn,
            vmin=vmin,
            vmax=vmax,
            spinbox_decimals=spinbox_decimals,
            unit=unit,
            si=si,
            **kwargs,
        )
        max_lq = self.New(
            name=f"{name}_max",
            initial=mx,
            vmin=vmin,
            vmax=vmax,
            spinbox_decimals=spinbox_decimals,
            unit=unit,
            si=si,
            **kwargs,
        )

        step_lq = self.New(
            f"{name}_step", initial=d, spinbox_decimals=spinbox_decimals + 1, **kwargs
        )
        step0 = int((initials[1] - initials[0]) / d) + 1
        num_lq = self.New(f"{name}_num", dtype=int, vmin=1, initial=step0, **kwargs)

        LQRange_kwargs = {
            "min_lq": min_lq,
            "max_lq": max_lq,
            "step_lq": step_lq,
            "num_lq": num_lq,
        }
        if include_center_span:
            c0 = (initials[1] + initials[0]) / 2
            s0 = initials[1] - initials[0]
            center_lq = self.New(
                name=f"{name}_center",
                initial=c0,
                spinbox_decimals=spinbox_decimals,
                unit=unit,
                si=si,
                **kwargs,
            )
            span_lq = self.New(
                name=f"{name}_span",
                initial=s0,
                spinbox_decimals=spinbox_decimals,
                unit=unit,
                si=si,
                **kwargs,
            )
            LQRange_kwargs.update({"center_lq": center_lq, "span_lq": span_lq})
        if include_sweep_type:
            sweep_type_lq = self.New(f"{name}_sweep_type", str, choices=("up", "down"))
            LQRange_kwargs.update({"sweep_type_lq": sweep_type_lq})

        lqrange = LQRange(**LQRange_kwargs)

        self.ranges[name] = lqrange
        return lqrange

    def New_Vector(self, name, components="xyz", initial=[1, 0, 0], **kwargs):

        assert len(components) == len(initial)

        if len(components) == 3:

            lq_x = self.New(f"{name}_{components[0]}", initial=initial[0], **kwargs)
            lq_y = self.New(f"{name}_{components[1]}", initial=initial[1], **kwargs)
            lq_z = self.New(f"{name}_{components[2]}", initial=initial[2], **kwargs)

            lq_vector = LQ3Vector(lq_x, lq_y, lq_z)
            self.vectors[name] = lq_vector
            return lq_vector

    def iter(self, include=None, exclude=None):
        """
        returns a list logged quantity specified by include and exclude

        include and exclude support wildcards:
            *       matches everything
            ?       matches any single character
            [seq]   matches any character in seq
            [!seq]  matches any char not in seq
        """
        return (
            (x, self.get_lq(x))
            for x in filter_with_patterns(self.keys(), include, exclude)
        )

    def New_UI(
        self,
        include=None,
        exclude=None,
        style="form",
        title=None,
    ) -> QtWidgets.QWidget:
        """
        create a default Qt Widget that contains
        widgets for all settings in the LQCollection
        """
        if title is not None:
            widget = QtWidgets.QGroupBox()
            widget.setTitle(title)
        else:
            widget = QtWidgets.QWidget()

        if style in ("form", "scroll_form"):
            layout = QtWidgets.QFormLayout(widget)

        elif style == "hbox":
            layout = QtWidgets.QHBoxLayout(widget)

        tools = Tools(layout, include, exclude)
        self._widgets_managers_.append(LQCollectionWidgetsManager(self, tools))

        if style == "scroll_form":
            scroll_area = QtWidgets.QScrollArea()
            scroll_area.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
            scroll_area.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
            scroll_area.setWidgetResizable(True)
            scroll_area.setWidget(widget)
            return scroll_area

        return widget

    def disconnect_all_from_hardware(self):
        for lq in self.as_list():
            lq.disconnect_from_hardware()


class LQCollectionWidgetsManager:

    def __init__(
        self,
        settings: LQCollection,
        tools: Tools,
    ) -> None:
        self.settings = settings
        self.tools = tools

        settings.q_object.lq_added.connect(self.add)
        settings.q_object.lq_removed.connect(self.remove)

        self.widgets = {}

        self.update()

    def add(self, lq: LoggedQuantity):
        self.update()

    def remove(self, lq: LoggedQuantity):
        widget = self.widgets.pop(lq.name, None)
        if widget is None:
            return
        self.tools.remove_from_layout(lq.name, widget)

    def update(self):
        for lqname, lq in tuple(
            self.settings.iter(self.tools.include, self.tools.exclude)
        ):
            if lqname in self.widgets.keys():
                continue
            if isinstance(lq, ArrayLQ):
                lineedit = QtWidgets.QLineEdit()
                button = QtWidgets.QPushButton("...")
                new_widget = QtWidgets.QWidget()
                layout = QtWidgets.QHBoxLayout()
                new_widget.setLayout(layout)
                layout.addWidget(lineedit)
                layout.addWidget(button)
                layout.setSpacing(0)
                layout.setContentsMargins(0, 0, 0, 0)
                lq.connect_to_widget(lineedit)
                button.clicked.connect(lq.array_tableView.show)
                button.clicked.connect(lq.array_tableView.raise_)
            elif isinstance(lq, FileLQ):
                new_widget = lq.new_default_widget()
                new_widget.layout().setSpacing(0)
                new_widget.layout().setContentsMargins(0, 0, 0, 0)
            elif isinstance(lq, LoggedQuantity):
                new_widget = lq.new_default_widget()
            else:
                continue

            self.widgets[lqname] = new_widget
            self.tools.add_to_layout(lqname, new_widget)



from typing import List, Protocol, Union

from qtpy import QtCore, QtWidgets

from ScopeFoundry.dynamical_widgets.tools import Tools
from ScopeFoundry.logged_quantity.collection import (
    LQCollection,
    LQCollectionWidgetsManager,
)
from ScopeFoundry.operations import Operations, OperationWidgetsManager


class Widgetable(Protocol):

    name: str
    settings: LQCollection
    operations: Operations
    _widgets_managers_: List


def new_widget(
    obj: Union[Widgetable, Operations, LQCollection],
    title: str = None,
    include=None,
    exclude=None,
    style="scroll_form",
) -> Union[QtWidgets.QScrollArea, QtWidgets.QWidget]:
    """returns a widget that represents objects with their settings and operations"""
    assert style in ("form", "hbox", "scroll_form")

    if title:
        widget = QtWidgets.QGroupBox()
        widget.setTitle(title)
    else:
        widget = QtWidgets.QWidget()

    if style in ("form", "scroll_form"):
        layout = QtWidgets.QFormLayout()

    elif style == "hbox":
        layout = QtWidgets.QHBoxLayout()

    if hasattr(obj, "settings") and hasattr(obj, "operations"):
        vlayout = QtWidgets.QVBoxLayout(widget)
        vlayout.addLayout(layout)
        add_to_layout(obj.settings, layout, include, exclude)
        add_to_layout(obj.operations, vlayout, include, exclude)
    else:
        widget.setLayout(layout)
        add_to_layout(obj, layout, include, exclude)

    if style == "scroll_form":
        scroll_area = QtWidgets.QScrollArea()
        scroll_area.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOn)
        scroll_area.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        scroll_area.setWidgetResizable(True)
        scroll_area.setWidget(widget)
        return scroll_area

    return widget


def add_to_layout(
    obj: Union[Widgetable, Operations, LQCollection],
    layout: Union[QtWidgets.QFormLayout, QtWidgets.QHBoxLayout, QtWidgets.QVBoxLayout],
    include=None,
    exclude=None,
) -> None:
    """dynamically adds widgets to a given layout."""

    tools = Tools(layout, include, exclude)

    if isinstance(obj, LQCollection):
        obj._widgets_managers_.append(LQCollectionWidgetsManager(obj, tools))
    elif isinstance(obj, Operations):
        obj._widgets_managers_.append(OperationWidgetsManager(obj, tools))
    else:
        obj._widgets_managers_.append(LQCollectionWidgetsManager(obj.settings, tools))
        obj._widgets_managers_.append(OperationWidgetsManager(obj.operations, tools))



import argparse
from collections import OrderedDict
from pathlib import Path

from qtpy import QtCore, QtWidgets, QtGui

from ScopeFoundry import BaseApp, LoggedQuantity
from ScopeFoundry.data_browser.data_browser_view import DataBrowserView
from ScopeFoundry.dynamical_widgets.tree_widget import new_tree_widget
from ScopeFoundry.helper_funcs import load_qt_ui_file, load_qt_ui_from_pkg, sibling_path
from ScopeFoundry.data_browser.viewers.file_info import FileInfoView


class DataBrowser(BaseApp):

    name = "DataBrowser"

    def __init__(self, argv, dark_mode=None):
        BaseApp.__init__(self, argv)

        self.views = OrderedDict()
        self.view_stack_indices = {}

        self.plugin_update_funcs = []
        self.plugins = {}
        self.keys = {
            QtCore.Qt.Key_Delete: self.on_recycle,
        }
        self.ctrl_keys = {
            QtCore.Qt.Key_R: self.on_rename,
        }

        s = self.settings
        s.New("data_filename", dtype="file")
        s.New("browse_dir", dtype="file", is_dir=True, initial="/")
        s.New("file_filter", dtype=str, initial="*.*,")
        s.New(
            "auto_select_view",
            dtype=bool,
            initial=True,
            description="auto selects the view when file name is changed.",
        )
        s.New("view_name", dtype=str, initial="file_info", choices=("0",))

        self.setup_ui()

        self.current_view = self.add_view(FileInfoView(self))
        self.setup_view(self.current_view)

        # add callbacks
        s.data_filename.add_listener(self.on_change_data_filename)
        s.data_filename.add_listener(self.on_change_data_filename_handle_plugins)

        s.browse_dir.add_listener(self.on_change_browse_dir)
        s["browse_dir"] = Path.cwd()

        s.file_filter.add_listener(self.on_change_file_filter)
        s.view_name.add_listener(self.on_change_view_name)

        self.setup_dark_mode_option(dark_mode)

        self.setup()

        # Does this even work?
        parser = argparse.ArgumentParser()
        for lq in self.settings.as_list():
            parser.add_argument(f"--{lq.name}")
        args = parser.parse_args()
        for lq in self.settings.as_list():
            if lq.name in args:
                val = getattr(args, lq.name)
                if val is not None:
                    lq.update_value(val)

    def setup(self):
        """overite in child to add further views and plug_ins"""
        pass

    def setup_ui(self):
        self.ui = load_qt_ui_from_pkg("ScopeFoundry.data_browser", "data_browser.ui")

        self.ui.setWindowTitle("ScopeFoundry: Data Browser")

        logo_path = sibling_path(__file__, "logo.png")
        self.set_logo(logo_path)

        # file system tree
        self.tree_view = self.ui.treeView
        self.fs_model = QtWidgets.QFileSystemModel()
        self.fs_model.setRootPath(QtCore.QDir.currentPath())
        self.tree_view.setModel(self.fs_model)
        self.tree_view.setIconSize(QtCore.QSize(16, 16))
        self.tree_view.setSortingEnabled(True)
        self.tree_view.setColumnWidth(0, 500)  # make name column wider
        self.tree_selectionModel = self.tree_view.selectionModel()
        self.tree_selectionModel.selectionChanged.connect(
            self.on_treeview_selection_change
        )

        self.view_stack: QtWidgets.QStackedWidget = self.ui.views_stack_widget

        # UI Connections
        s = self.settings
        s.data_filename.connect_to_browse_widgets(
            self.ui.data_filename_lineEdit, self.ui.data_filename_browse_pushButton
        )
        s.browse_dir.connect_to_browse_widgets(
            self.ui.browse_dir_lineEdit, self.ui.browse_dir_browse_pushButton
        )
        self.ui.data_filename_recycle_pushButton.clicked.connect(self.on_recycle)
        self.ui.data_filename_rename_pushButton.clicked.connect(self.on_rename)
        s.view_name.connect_to_widget(self.ui.view_name_comboBox)
        s.auto_select_view.connect_to_widget(self.ui.auto_select_checkBox)
        s.file_filter.connect_to_widget(self.ui.file_filter_lineEdit)

        self.ui.console_pushButton.clicked.connect(self.console_widget.show)
        self.ui.log_pushButton.clicked.connect(self.logging_widget.show)

        self.ui.action_save_as_ini.triggered.connect(self.settings_save_dialog)
        self.ui.action_load_ini.triggered.connect(self.settings_load_dialog)
        self.ui.action_show_settings.triggered.connect(self.on_show_settings)
        self.ui.show_settings_pushButton.clicked.connect(self.on_show_settings)

        self.ui.show()
        self.ui.raise_()

        self.ui.keyPressEvent = self.handle_key_board

    def set_logo(self, logo_path):
        logo_icon = QtGui.QIcon(logo_path)
        self.qtapp.setWindowIcon(logo_icon)
        self.ui.setWindowIcon(logo_icon)

    def handle_key_board(self, event):
        if event.modifiers() != QtCore.Qt.ControlModifier:
            func = self.keys.get(event.key(), None)
            if func is not None:
                func()
            return

        func = self.ctrl_keys.get(event.key(), None)
        if func is not None:
            func()

    def add_view(self, new_view):
        self.log.debug(f"add_view called {new_view}")

        # Update views dict
        self.views[new_view.name] = new_view

        self.add_lq_collection_to_settings_path(new_view.settings)

        self.settings.get_lq("view_name").change_choice_list(list(self.views.keys()))

        self.log.debug(f"add_view done {new_view}")

        return new_view

    def add_plugin(self, plugin):
        self.ui.plugin_buttons_layout.addWidget(plugin.get_show_hide_button())
        self.plugin_update_funcs.append(plugin.update_if_showing)
        self.ctrl_keys[plugin.show_keyboard_key] = plugin.toggle_show_hide
        self.plugins[plugin.name] = plugin
        self.add_lq_collection_to_settings_path(plugin.settings)

    def load_view(self, new_view):
        # deprecated use DataBrowser.add_view(new_view) instead
        self.add_view(new_view)

    def on_change_data_filename(self):
        fname = self.settings["data_filename"]
        if not Path(fname).is_file():
            print("invalid data_filename", fname)
            return

        print("file", fname)

        # select view
        if self.settings["auto_select_view"]:
            view_name = self.auto_select_view(fname)
            if view_name != self.current_view.name:
                self.settings["view_name"] = view_name
                self.setup_view(self.views[view_name])

        self.current_view.on_change_data_filename(fname)

    def on_change_data_filename_handle_plugins(self):
        fname = self.settings["data_filename"]
        for func in self.plugin_update_funcs:
            func(fname)

    @QtCore.Slot()
    def on_change_browse_dir(self):
        self.log.debug("on_change_browse_dir")
        self.ui.treeView.setRootIndex(self.fs_model.index(self.settings["browse_dir"]))
        self.fs_model.setRootPath(self.settings["browse_dir"])

    def on_change_file_filter(self):
        self.log.debug("on_change_file_filter")
        filter_str = self.settings["file_filter"]
        if filter_str == "":
            filter_str = "*"
            self.settings["file_filter"] = "*"
        filter_str_list = [x.strip() for x in filter_str.split(",")]
        self.log.debug(filter_str_list)
        self.fs_model.setNameFilters(filter_str_list)

    def setup_view(self, view: DataBrowserView):

        if not view.view_loaded:
            view.setup()
            index = self.view_stack.addWidget(self.current_view.ui)
            self.view_stack_indices[view.name] = index
            view.view_loaded = True

        self.current_view = view
        self.view_stack.setCurrentIndex(self.view_stack_indices[view.name])

    def on_change_view_name(self):
        self.current_view = self.views[self.settings["view_name"]]
        self.setup_view(self.current_view)

        # udpate
        fname = self.settings["data_filename"]
        if Path(fname).is_file():
            self.current_view.on_change_data_filename(fname)

    def on_treeview_selection_change(self, sel, desel):
        fname = self.fs_model.filePath(self.tree_selectionModel.currentIndex())
        self.settings["data_filename"] = fname

    #        print( 'on_treeview_selection_change' , fname, sel, desel)

    def auto_select_view(self, fname):
        "return the name of the last supported view for the given fname"
        for view_name, view in list(self.views.items())[::-1]:
            if view.is_file_supported(fname):
                return view_name
        return "file_info"

    def on_recycle(self):
        # import here because send2trash is not a built in library.
        import send2trash

        send2trash.send2trash(Path(self.settings["data_filename"]))

    def on_rename(self):
        fname = self.settings["data_filename"]
        dialog = RenameDialog(prev_path_name=fname)
        if not dialog.exec_():
            return  # dialog was escaped

        if dialog.new_name:
            Path(fname).rename(dialog.new_name)

    def read_setting(self, path, ini_string_value=False):
        lq = self.get_lq(path)

        if ini_string_value:
            return lq.ini_string_value()
        return lq.val

    def get_lq(self, path: str) -> LoggedQuantity:
        """
        returns the LoggedQuantity defined by a path string of the form 'section/[component/]setting'
        where section are "view", "plugin" or "app"
        returns None if it does not exist
        """
        parts = path.split("/")

        settings = self.settings
        if parts[0] == "view":
            view = self.views.get(parts[1], None)
            if view is not None:
                self.load_view(view)
                settings = view.settings
        elif parts[0] == "plugin":
            plugin = self.plugins.get(parts[1], None)
            if plugin is not None:
                settings = plugin.settings

        if parts[-1] in settings:
            return settings.get_lq(parts[-1])

    def settings_save_dialog(self):
        """Opens a save as ini dialogue in the app user interface."""
        fname, _ = QtWidgets.QFileDialog.getSaveFileName(
            self.ui, "Save Settings file", "", "Settings File (*.ini)"
        )
        if fname:
            self.settings_save_ini(fname)

    def settings_load_dialog(self):
        """Opens a load ini dialogue in the app user interface"""
        fname, _ = QtWidgets.QFileDialog.getOpenFileName(
            self.ui, "Open Settings file", "", "Settings File (*.ini)"
        )
        if fname.endswith(".ini"):
            self.settings_load_ini(fname)
        # elif fname.endswith(".h5"):
        #     self.settings_load_h5(fname)

    def on_show_settings(self):
        if not hasattr(self, "_trees_dialog"):

            app_tree = new_tree_widget((self,), ("app", ""))
            views_tree = new_tree_widget(
                [obj for obj in self.views.values()], ("views", "")
            )
            plugins_tree = new_tree_widget(
                [obj for obj in self.plugins.values()], ("plugins", "")
            )

            splitter = QtWidgets.QSplitter()
            splitter.addWidget(app_tree)
            splitter.addWidget(views_tree)
            splitter.addWidget(plugins_tree)

            pb_hlayout = QtWidgets.QHBoxLayout()
            pb = QtWidgets.QPushButton("save ini ...")
            pb.clicked.connect(self.settings_save_ini_ask)
            pb_hlayout.addWidget(pb)
            pb = QtWidgets.QPushButton("load ini ...")
            pb.clicked.connect(self.settings_load_ini_ask)
            pb_hlayout.addWidget(pb)

            layout = QtWidgets.QVBoxLayout()
            layout.addWidget(splitter)
            layout.addLayout(pb_hlayout)

            self._trees_dialog = QtWidgets.QDialog()
            self._trees_dialog.setLayout(layout)

        self._trees_dialog.exec_()


class RenameDialog(QtWidgets.QDialog):

    def __init__(self, prev_path_name):
        QtWidgets.QDialog.__init__(self)

        self.new_name = None

        rename_btn = QtWidgets.QPushButton("rename")
        cancel_btn = QtWidgets.QPushButton("cancel")
        self.new_name_w = QtWidgets.QLineEdit(prev_path_name)

        cancel_btn.clicked.connect(self.on_cancel)
        rename_btn.clicked.connect(self.on_rename)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.new_name_w)
        layout.addWidget(rename_btn)
        layout.addWidget(cancel_btn)
        self.setLayout(layout)

        self.setWindowTitle("rename")
        self.setMinimumWidth(500)

    def on_cancel(self):
        self.new_name = None
        self.accept()

    def on_rename(self):
        self.new_name = self.new_name_w.text()
        self.accept()



from functools import partial
from typing import Protocol, List, Union

from qtpy import QtCore, QtGui, QtWidgets

from ScopeFoundry.helper_funcs import filter_with_patterns
from ScopeFoundry.logged_quantity import ArrayLQ, FileLQ, LoggedQuantity, LQCollection
from ScopeFoundry.operations import Operations


class SubtreeAbleObj(Protocol):

    name: str
    settings: LQCollection
    operations: Operations
    _subtree_managers_: List

    def on_new_subtree(self, subtree): ...  # optional

    def on_right_click(self): ...  # optional


def new_tree_widget(
    objs: List[SubtreeAbleObj], header=["col0", ""], include=None, exclude=None
) -> QtWidgets.QTreeWidget:
    """returns a tree widget that represents objects with their settings and operations"""
    tree_widget = QtWidgets.QTreeWidget()
    tree_widget.setColumnCount(len(header))
    tree_widget.setHeaderLabels(header)
    tree_widget.setColumnWidth(0, 180)
    tree_widget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
    tree_widget.customContextMenuRequested.connect(
        partial(on_right_click, tree=tree_widget)
    )

    for obj in objs:
        SubtreeManager(tree_widget, obj, include, exclude)

    return tree_widget


class SFQTreeWidgetItem(QtWidgets.QTreeWidgetItem):

    obj: Union[SubtreeAbleObj, LoggedQuantity, Operations]


class SubtreeManager:
    """
    adds tree items to the *tree_widget* forming.

     *tree_widget*
        *header_item* (obj dependent)
            - child_0 (LoggedQuantity or OperationBtn)
            - child_1 (LoggedQuantity or OperationBtn)
           ...
        ...

    the
    """

    def __init__(
        self,
        tree_widget: QtWidgets.QTreeWidget,
        obj: SubtreeAbleObj,
        include=None,
        exclude=None,
    ) -> None:
        self.name = obj.name
        self.settings = obj.settings
        self.operations = obj.operations
        self.include = include
        self.exclude = exclude
        self.settings.q_object.lq_added.connect(self.add_lq_child_item)
        self.settings.q_object.lq_removed.connect(self.remove_lq_child_item)
        self.operations.q_object.added.connect(self.add_operation_child_item)
        self.operations.q_object.removed.connect(self.remove_operation_child_item)

        self.settings_items = {}
        self.operation_items = {}

        self.header_item = SFQTreeWidgetItem(tree_widget, [obj.name, ""])
        self.tree_widget = tree_widget  # the widget where instance lives

        if hasattr(obj, "on_new_subtree"):
            obj.on_new_subtree(self)
        obj._subtree_managers_.append(self)

        tree_widget.insertTopLevelItem(0, self.header_item)
        update_settings(
            self.settings,
            self.header_item,
            self.settings_items,
            self.include,
            self.exclude,
        )
        update_operations(
            self.operations,
            self.header_item,
            self.operation_items,
            self.include,
            self.exclude,
        )
        self.header_item.obj = obj

    def add_lq_child_item(self, lq: LoggedQuantity):
        update_settings(
            self.settings,
            self.header_item,
            self.settings_items,
            self.include,
            self.exclude,
        )

    def remove_lq_child_item(self, lq: LoggedQuantity):
        remove_from_tree(lq.name, self.header_item, self.settings_items)

    def add_operation_child_item(self, name: str):
        update_operations(
            self.operations,
            self.header_item,
            self.operation_items,
            self.include,
            self.exclude,
        )

    def remove_operation_child_item(self, name: str):
        remove_from_tree(name, self.header_item, self.operation_items)

    def set_header_text(self, col=1, text="", color=None):
        self.header_item.setText(col, text)
        if color is not None:
            self.header_item.setForeground(col, QtGui.QColor(color))


def remove_from_tree(
    name: str,
    root_item: SFQTreeWidgetItem,
    children,  # dict[str, SFQTreeWidgetItem],
) -> None:
    item = children.pop(name, None)
    if item is None:
        return
    root_item.removeChild(item)


def update_operations(
    operations: Operations,
    root_item: QtWidgets.QTreeWidgetItem,
    children,  #: dict[str, QtWidgets.QTreeWidgetItem],
    include,
    exclude,
) -> None:
    for op_name in filter_with_patterns(operations.keys(), include, exclude):
        if op_name in children.keys():
            continue
        op_button = operations.new_button(op_name)
        new_item = QtWidgets.QTreeWidgetItem(root_item, [op_name, ""])
        children[op_name] = new_item
        root_item.addChild(new_item)
        root_item.treeWidget().setItemWidget(new_item, 1, op_button)


def update_settings(
    settings: LQCollection,
    root_item: SFQTreeWidgetItem,
    children,  #: dict[str, SFQTreeWidgetItem],
    include=None,
    exclude=None,
) -> None:
    for lqname, lq in tuple(settings.iter(include, exclude)):
        if lqname in children.keys():
            continue
        if isinstance(lq, ArrayLQ):
            lineedit = QtWidgets.QLineEdit()
            button = QtWidgets.QPushButton("...")
            widget = QtWidgets.QWidget()
            layout = QtWidgets.QHBoxLayout()
            widget.setLayout(layout)
            layout.addWidget(lineedit)
            layout.addWidget(button)
            layout.setSpacing(0)
            layout.setContentsMargins(0, 0, 0, 0)
            lq.connect_to_widget(lineedit)
            button.clicked.connect(lq.array_tableView.show)
            button.clicked.connect(lq.array_tableView.raise_)
        elif isinstance(lq, FileLQ):
            widget = lq.new_default_widget()
            widget.layout().setSpacing(0)
            widget.layout().setContentsMargins(0, 0, 0, 0)
        elif isinstance(lq, LoggedQuantity):
            widget = lq.new_default_widget()
        else:
            continue

        new_item = SFQTreeWidgetItem(root_item, [lqname, ""])
        new_item.obj = lq

        children[lqname] = new_item
        root_item.addChild(new_item)
        root_item.treeWidget().setItemWidget(new_item, 1, widget)


def on_right_click(position, tree: QtWidgets.QTreeWidget) -> None:
    selected_items = tree.selectedItems()
    if len(selected_items) < 1:
        return
    selected_item: SFQTreeWidgetItem = selected_items[0]
    if hasattr(selected_item.obj, "on_right_click"):
        selected_item.obj.on_right_click()



import operator
import time
from typing import TypedDict

from qtpy.QtWidgets import QComboBox, QLineEdit

from .editor_base_ui import EditorBaseUI
from .base_item import BaseItem
from .helper_func import new_q_completer


ITEM_TYPE = "wait-until"
DESCRIPTION = "sequencer sleeps until a condition is met"


class InterruptIfKwargs(TypedDict):
    setting: str
    operator: str
    value: str


OPERATORS = {
    "=": operator.eq,
    ">": operator.gt,
    "<": operator.lt,
    ">=": operator.ge,
    "<=": operator.le,
}

class WaitUntil(BaseItem):
    item_type = ITEM_TYPE

    def visit(self):
        relate = OPERATORS[self.kwargs["operator"]]
        lq = self.app.get_lq(self.kwargs["setting"])
        val = lq.coerce_to_type(self.kwargs["value"])
        while True:
            if relate(lq.val, val) or self.measure.interrupt_measurement_called:
                break
            time.sleep(0.05)


class WaitUntilEditorUI(EditorBaseUI):
    item_type = ITEM_TYPE
    description = DESCRIPTION

    def __init__(self, measure, paths) -> None:
        self.paths = paths
        super().__init__(measure)

    def setup_ui(self):
        self.setting_cb = QComboBox()
        self.setting_cb.setEditable(True)
        self.setting_cb.addItems(self.paths)
        self.setting_cb.setToolTip("setting")
        self.setting_cb.setCompleter(new_q_completer(self.paths))
        self.layout.addWidget(self.setting_cb)
        self.operator_cb = QComboBox()
        self.operator_cb.addItems(list(OPERATORS.keys()))
        self.layout.addWidget(self.operator_cb)
        self.value_le = QLineEdit()
        self.value_le.setToolTip("wait until setting reaches this value")
        self.layout.addWidget(self.value_le)

    def get_kwargs(self) -> InterruptIfKwargs:
        path = self.setting_cb.currentText()
        o = self.operator_cb.currentText()
        v = self.value_le.text()
        return {"setting": path, "operator": o, "value": v}

    def set_kwargs(self, **kwargs):
        self.setting_cb.setCurrentText(kwargs["setting"])
        self.operator_cb.setCurrentText(kwargs["operator"])
        self.value_le.setText(kwargs["value"])
        self.value_le.selectAll()
        self.value_le.setFocus()



from pathlib import Path

from qtpy import QtWidgets

from .base_app import WRITE_RES
from ..helper_funcs import open_file


def show_io_report_dialog(fname: str, report, retry_func):

    failures = [f" - `{p}`" for p, v in report.items() if v == WRITE_RES.MISSING]
    protecteds = [f" - `{p}`" for p, v in report.items() if v == WRITE_RES.PROTECTED]
    successes = [f" - `{p}`" for p, v in report.items() if v is WRITE_RES.SUCCESS]

    name = Path(fname).name

    if not failures:
        if protecteds:
            print(f"{name} has protected settings that were not updated:")
            for p in protecteds:
                print(p)
        return

    lines = [f"### failed ({len(failures)})", "*check typos*", " "] + failures

    if protecteds:
        lines += [
            "  ",
            "  ",
            f"### protected ({len(protecteds)})",
            "*`protected` settings can not be update by a file*",
            "  ",
        ]
        lines += protecteds

    if successes:
        lines += ["  ", "  ", f"### sucesses ({len(successes)})"]
        lines += successes

    text_edit = QtWidgets.QTextEdit()
    text_edit.setMarkdown("\n".join(lines))

    dialog = QtWidgets.QDialog()
    dialog.setWindowTitle(f"{name} failed to update ({len(failures)})")
    layout = QtWidgets.QVBoxLayout()
    dialog.setLayout(layout)
    layout.addWidget(text_edit)

    if name.endswith(".ini"):

        def on_fix():
            open_file(fname)

        btn = QtWidgets.QPushButton("fix")
        btn.clicked.connect(on_fix)
        layout.addWidget(btn)

        def on_retry():
            dialog.close()
            retry_func(fname)

        btn = QtWidgets.QPushButton("retry")
        btn.clicked.connect(on_retry)
        layout.addWidget(btn)

    btn = QtWidgets.QPushButton(f"close")
    btn.clicked.connect(dialog.close)
    layout.addWidget(btn)

    dialog.exec_()



import unittest

from ScopeFoundry import BaseMicroscopeApp, HardwareComponent, Measurement


class Measure1(Measurement):
    name = "measure1"

    def setup(self):
        self.settings.New("string", str, ro=True, initial="0")


class Hardware1(HardwareComponent):
    name = "hardware1"

    def setup(self):
        self.settings.New("string", str, ro=True, initial="0")
        self.settings.New("float", float, ro=True, initial=0.0)
        self.settings.New("int", float, ro=True, initial=0.0)

    def connect(self):
        self.settings.get_lq("string").connect_to_hardware(print, None)
        self.settings.get_lq("float").connect_to_hardware(print, print)
        self.settings.get_lq("int").connect_to_hardware(None, print)

    def disconnect(self):
        self.settings.disconnect_all_from_hardware()


class App(BaseMicroscopeApp):

    mdi = False

    def setup(self):
        self.add_measurement(Measure1(self))
        self.hw = self.add_hardware(Hardware1(self))

class GetSettingsTest(unittest.TestCase):

    def setUp(self):
        self.app = App([])
        self.hw = self.app.hw

    def test_settings_persistance(self):
        all_settings_0 = list(self.app.get_setting_paths())
        self.hw.connect()
        self.assertListEqual(all_settings_0, list(self.app.get_setting_paths()))
        self.hw.disconnect()
        self.assertListEqual(all_settings_0, list(self.app.get_setting_paths()))

    def test_filter_has_hardware_read(self):
        self.hw.connect()
        self.assertListEqual(
            ["hw/hardware1/string", "hw/hardware1/float"],
            list(self.app.get_setting_paths(filter_has_hardware_read=True)),
        )
        self.hw.disconnect()
        self.assertListEqual(
            [], list(self.app.get_setting_paths(filter_has_hardware_read=True))
        )

    def test_filter_has_hardware_write(self):
        self.hw.connect()
        self.assertListEqual(
            ["hw/hardware1/float", "hw/hardware1/int"],
            list(self.app.get_setting_paths(filter_has_hardware_write=True)),
        )
        self.hw.disconnect()
        self.assertListEqual(
            [], list(self.app.get_setting_paths(filter_has_hardware_write=True))
        )

    def test_filter_has_hardware_read_and_write(self):
        self.hw.connect()
        self.assertListEqual(
            ["hw/hardware1/float", "hw/hardware1/int"],
            list(self.app.get_setting_paths(filter_has_hardware_write=True)),
        )
        self.hw.disconnect()
        self.assertListEqual(
            [], list(self.app.get_setting_paths(filter_has_hardware_write=True))
        )

    def test_ro(self):
        self.assertSetEqual(
            {
                "mm/measure1/activation",
                "mm/measure1/profile",
                "hw/hardware1/debug_mode",
                "hw/hardware1/connected",
            },
            set(
                self.app.get_setting_paths(exclude_ro=True, exclude_patterns=("app/*",))
            ),
        )

    def test_pattern(self):
        self.assertSetEqual(
            {
                "hw/hardware1/debug_mode",
                "hw/hardware1/int",
                "hw/hardware1/float",
                "hw/hardware1/string",
            },
            set(
                self.app.get_setting_paths(
                    exclude_patterns=("mm/*", "app/*", "*connected")
                )
            ),
        )


if __name__ == "__main__":
    unittest.main()



# Form implementation generated from reading ui file 'base_microscope_app_mdi.ui'
#
# Created by: PyQt6 UI code generator 6.7.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from qtpy import QtCore, QtGui, QtWidgets


class Ui_MainWindow(QtWidgets.QMainWindow):

    def setupUi(self, MainWindow: QtWidgets.QMainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1105, 797)
        icon = QtGui.QIcon()
        icon.addPixmap(
            QtGui.QPixmap("scopefoundry_logo2_1024.png"),
            QtGui.QIcon.Mode.Normal,
            QtGui.QIcon.State.Off,
        )
        MainWindow.setWindowIcon(icon)
        MainWindow.setUnifiedTitleAndToolBarOnMac(False)
        self.centralwidget = QtWidgets.QWidget(parent=MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.col_splitter = QtWidgets.QSplitter(parent=self.centralwidget)
        self.col_splitter.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.col_splitter.setHandleWidth(20)
        self.col_splitter.setChildrenCollapsible(True)
        self.col_splitter.setObjectName("col_splitter")
        self.quickaccess_scrollArea = QtWidgets.QScrollArea(parent=self.col_splitter)
        self.quickaccess_scrollArea.setMinimumSize(QtCore.QSize(400, 0))
        self.quickaccess_scrollArea.setWidgetResizable(True)
        self.quickaccess_scrollArea.setObjectName("quickaccess_scrollArea")
        self.quickaccess_scrollAreaWidgetContents = QtWidgets.QWidget()
        self.quickaccess_scrollAreaWidgetContents.setGeometry(
            QtCore.QRect(0, 0, 398, 744)
        )
        self.quickaccess_scrollAreaWidgetContents.setObjectName(
            "quickaccess_scrollAreaWidgetContents"
        )
        self.quickaccess_layout = QtWidgets.QVBoxLayout(
            self.quickaccess_scrollAreaWidgetContents
        )
        self.quickaccess_layout.setObjectName("quickaccess_layout")
        self.quickaccess_scrollArea.setWidget(self.quickaccess_scrollAreaWidgetContents)
        self.lq_trees_widget = QtWidgets.QWidget(parent=self.col_splitter)
        self.lq_trees_widget.setMinimumSize(QtCore.QSize(325, 0))
        self.lq_trees_widget.setObjectName("lq_trees_widget")
        self.tree_layout = QtWidgets.QVBoxLayout(self.lq_trees_widget)
        self.tree_layout.setContentsMargins(0, 0, 0, 0)
        self.tree_layout.setObjectName("tree_layout")
        self.mdiArea = QtWidgets.QMdiArea(parent=self.col_splitter)
        self.mdiArea.setViewMode(QtWidgets.QMdiArea.ViewMode.TabbedView)
        self.mdiArea.setDocumentMode(False)
        self.mdiArea.setTabsMovable(True)
        self.mdiArea.setObjectName("mdiArea")
        self.gridLayout_2.addWidget(self.col_splitter, 0, 0, 1, 1)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(parent=MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1105, 17))
        self.menubar.setNativeMenuBar(False)
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(parent=self.menubar)
        self.menuFile.setObjectName("menuFile")
        self.menuAdvanced = QtWidgets.QMenu(parent=self.menubar)
        self.menuAdvanced.setObjectName("menuAdvanced")
        self.menuHelp = QtWidgets.QMenu(parent=self.menubar)
        self.menuHelp.setObjectName("menuHelp")
        self.menuWindow = QtWidgets.QMenu(parent=self.menubar)
        self.menuWindow.setObjectName("menuWindow")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(parent=MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.action_tab_mode = QtGui.QAction(parent=MainWindow)
        self.action_tab_mode.setCheckable(True)
        self.action_tab_mode.setChecked(True)
        self.action_tab_mode.setObjectName("action_tab_mode")
        self.action_set_data_dir = QtGui.QAction(parent=MainWindow)
        self.action_set_data_dir.setEnabled(True)
        self.action_set_data_dir.setObjectName("action_set_data_dir")
        self.action_load_ini = QtGui.QAction(parent=MainWindow)
        self.action_load_ini.setObjectName("action_load_ini")
        self.action_save_ini = QtGui.QAction(parent=MainWindow)
        self.action_save_ini.setObjectName("action_save_ini")
        self.action_about = QtGui.QAction(parent=MainWindow)
        self.action_about.setObjectName("action_about")
        self.action_console = QtGui.QAction(parent=MainWindow)
        self.action_console.setObjectName("action_console")
        self.action_docs = QtGui.QAction(parent=MainWindow)
        self.action_docs.setObjectName("action_docs")
        self.action_log_viewer = QtGui.QAction(parent=MainWindow)
        self.action_log_viewer.setEnabled(True)
        self.action_log_viewer.setObjectName("action_log_viewer")
        self.action_subwindow_mode = QtGui.QAction(parent=MainWindow)
        self.action_subwindow_mode.setCheckable(True)
        self.action_subwindow_mode.setObjectName("action_subwindow_mode")
        self.window_action = QtGui.QAction(parent=MainWindow)
        self.window_action.setCheckable(True)
        self.window_action.setEnabled(True)
        self.window_action.setObjectName("window_action")
        self.tab_action = QtGui.QAction(parent=MainWindow)
        self.tab_action.setCheckable(True)
        self.tab_action.setObjectName("tab_action")
        self.cascade_action = QtGui.QAction(parent=MainWindow)
        self.cascade_action.setObjectName("cascade_action")
        self.tile_action = QtGui.QAction(parent=MainWindow)
        self.tile_action.setObjectName("tile_action")
        self.action_auto_save_ini = QtGui.QAction(parent=MainWindow)
        self.action_auto_save_ini.setObjectName("action_auto_save_ini")
        self.action_save_window_positions = QtGui.QAction(parent=MainWindow)
        self.action_save_window_positions.setObjectName("action_save_window_positions")
        self.action_load_window_positions = QtGui.QAction(parent=MainWindow)
        self.action_load_window_positions.setObjectName("action_load_window_positions")
        self.action_load_last = QtGui.QAction(parent=MainWindow)
        self.action_load_last.setObjectName("action_load_last")
        self.action_load_all_measure_uis = QtGui.QAction(parent=MainWindow)
        self.action_load_all_measure_uis.setObjectName("action_load_all_measure_uis")
        self.action_show_settings = QtGui.QAction(parent=MainWindow)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(
            QtGui.QPixmap("settings.png"),
            QtGui.QIcon.Mode.Normal,
            QtGui.QIcon.State.Off,
        )
        self.action_show_settings.setIcon(icon1)
        self.action_show_settings.setObjectName("action_show_settings")
        self.action_propose_from_file = QtGui.QAction(parent=MainWindow)
        self.action_propose_from_file.setObjectName("action_propose_from_file")
        self.action_analyze_with_ipynb = QtGui.QAction(parent=MainWindow)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(
            QtGui.QPixmap("Jupyter_logo.png"),
            QtGui.QIcon.Mode.Normal,
            QtGui.QIcon.State.Off,
        )
        self.action_analyze_with_ipynb.setIcon(icon2)
        self.action_analyze_with_ipynb.setObjectName("action_analyze_with_ipynb")
        self.menuFile.addAction(self.action_set_data_dir)
        self.menuFile.addSeparator()
        self.menuFile.addAction(self.action_load_ini)
        self.menuFile.addAction(self.action_auto_save_ini)
        self.menuFile.addAction(self.action_save_ini)
        self.menuFile.addAction(self.action_load_last)
        self.menuAdvanced.addAction(self.action_console)
        self.menuAdvanced.addAction(self.action_log_viewer)
        self.menuAdvanced.addSeparator()
        self.menuAdvanced.addAction(self.action_show_settings)
        self.menuAdvanced.addAction(self.action_propose_from_file)
        self.menuAdvanced.addAction(self.action_analyze_with_ipynb)
        self.menuHelp.addAction(self.action_docs)
        self.menuHelp.addSeparator()
        self.menuHelp.addAction(self.action_about)
        self.menuWindow.addAction(self.action_save_window_positions)
        self.menuWindow.addAction(self.action_load_window_positions)
        self.menuWindow.addSeparator()
        self.menuWindow.addAction(self.window_action)
        self.menuWindow.addAction(self.tab_action)
        self.menuWindow.addAction(self.cascade_action)
        self.menuWindow.addAction(self.tile_action)
        self.menuWindow.addSeparator()
        self.menuWindow.addAction(self.action_load_all_measure_uis)
        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuAdvanced.menuAction())
        self.menubar.addAction(self.menuWindow.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow: QtWidgets.QMainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "ScopeFoundry"))
        self.menuFile.setTitle(_translate("MainWindow", "File"))
        self.menuAdvanced.setTitle(_translate("MainWindow", "Advanced"))
        self.menuHelp.setTitle(_translate("MainWindow", "Help"))
        self.menuWindow.setTitle(_translate("MainWindow", "Window"))
        self.action_tab_mode.setText(_translate("MainWindow", "Tab Mode"))
        self.action_tab_mode.setShortcut(_translate("MainWindow", "Alt+T"))
        self.action_set_data_dir.setText(
            _translate("MainWindow", "Set Data Save Directory ...")
        )
        self.action_set_data_dir.setShortcut(_translate("MainWindow", "Ctrl+D"))
        self.action_load_ini.setText(_translate("MainWindow", "&Load Settings ..."))
        self.action_load_ini.setShortcut(_translate("MainWindow", "Ctrl+O"))
        self.action_save_ini.setText(_translate("MainWindow", "&Save Settings As ..."))
        self.action_save_ini.setShortcut(_translate("MainWindow", "Ctrl+Shift+S"))
        self.action_about.setText(_translate("MainWindow", "&About"))
        self.action_console.setText(_translate("MainWindow", "&Console "))
        self.action_console.setShortcut(_translate("MainWindow", "Alt+C"))
        self.action_docs.setText(_translate("MainWindow", "ScopeFoundry Docs"))
        self.action_docs.setShortcut(_translate("MainWindow", "Alt+F"))
        self.action_log_viewer.setText(_translate("MainWindow", "Log Viewer"))
        self.action_log_viewer.setShortcut(_translate("MainWindow", "Alt+L"))
        self.action_subwindow_mode.setText(_translate("MainWindow", "Subwindow Mode"))
        self.window_action.setText(_translate("MainWindow", "Sub&window Mode"))
        self.window_action.setShortcut(_translate("MainWindow", "Alt+W"))
        self.tab_action.setText(_translate("MainWindow", "&Tab Mode"))
        self.tab_action.setShortcut(_translate("MainWindow", "Alt+T"))
        self.cascade_action.setText(_translate("MainWindow", "Cascade"))
        self.tile_action.setText(_translate("MainWindow", "Tile"))
        self.action_auto_save_ini.setText(_translate("MainWindow", "Save Settings"))
        self.action_auto_save_ini.setShortcut(_translate("MainWindow", "Ctrl+S"))
        self.action_save_window_positions.setText(
            _translate("MainWindow", "Save Window Positions ...")
        )
        self.action_load_window_positions.setText(
            _translate("MainWindow", "Load Window Positions ...")
        )
        self.action_load_last.setText(_translate("MainWindow", "Load last Settings"))
        self.action_load_all_measure_uis.setText(_translate("MainWindow", "load"))
        self.action_show_settings.setText(_translate("MainWindow", "Settings"))
        self.action_propose_from_file.setText(_translate("MainWindow", "Inspect file"))
        self.action_analyze_with_ipynb.setText(
            _translate("MainWindow", "analyze with ipynb")
        )



import random


class SimulonXYZStageDev:

    def __init__(self, debug=False):
        self.x = 0
        self.y = 0
        self.z = 0
        self.debug = debug
        # communicate with hardware here

    def read_x(self):
        self.x = self.x + self.noise()
        return self.x

    def read_y(self):
        self.y = self.y + self.noise()
        return self.y

    def read_z(self):
        self.z = self.z + self.noise()
        return self.z

    def write_x(self, x):
        self.x = x

    def write_y(self, y):
        self.y = y

    def write_z(self, z):
        self.z = z

    def close(self):
        pass

    def noise(self):
        return (random.random() - 0.5) * 10e-3
